/*
 * logging functions
 */

#include <log.h>

#include <stdio.h>
#ifdef __STDC__
# include <stdarg.h>
#else
# include <varargs.h>
#endif
#include <unistd.h>
#include <limits.h>
#include <signal.h>

static short	_log_syslog_init = 0;	/* == 1 after start_syslog() success */

static short	_tty = 1;		/* == value of isatty */

int		debug = 0;		/* global debug switch */
short		_debug_depth = 0;	/* stack depth BLECH! */

extern char	*progname;

/*
 *    function:	void debug_msg()
 * description:	print a debug message to stderr
 *     returns:	
 */
void
debug_msg (char *fmt, ...) {
    va_list		ap;
    char		buf[BUF_SZ];
    char		spaces[4];
    int			ret;

    sprintf(spaces, "%%%ds", _debug_depth);
    fprintf(stderr, spaces, "");

    va_start(ap, fmt);
    ret = vsnprintf(buf, BUF_SZ, fmt, ap);
    va_end(ap);

    /* read the string and replace any %'s with %% */
    replace_pc(buf);

    fprintf(stderr, buf);

    /* did the string exceed BUF_SZ? */
    if (ret > BUF_SZ - 1)
	if (_tty)
	    fprintf(stderr, "WARNING: last debug message exceeded buffer size:"
		" %d\n", BUF_SZ -1);

    return;
}

/*
 *    function:	void log_msg()
 * description:	print a debug message to stderr and syslog
 *     returns:	
 */
void
log_msg(int level, char *fmt, ...) {
    va_list		ap;
    char		buf[BUF_SZ];
    int			ret;

    va_start(ap, fmt);
    ret = vsnprintf(buf, BUF_SZ, fmt, ap);
    va_end(ap);

    if (_tty)
	fprintf(stderr, "%s: %s", progname, buf);

    /* read the string and replace any %'s with %% */
    replace_pc(buf);

    if (_log_syslog_init)
	syslog(level, buf);

    /* did the string exceed BUF_SZ? */
    if (ret > BUF_SZ - 1) {
	if (_tty)
	    fprintf(stderr, "WARNING: last log message exceeded log buffer "
		"size: %d\n", BUF_SZ - 1);
	if (_log_syslog_init)
	    syslog(LOG_ERR, "WARNING: last log message exceeded log buffer "
		"size: %d\n", BUF_SZ - 1);
    }

    return;
}

/*
 *    function:	void log_msgx()
 * description:	print a debug message to stderr and syslog, then exit
 *     returns:	
 */
void
log_msgx(int status, char *fmt, ...) {
    va_list		ap;
    char		buf[BUF_SZ];
    int			ret;

    va_start(ap, fmt);
    ret = vsnprintf(buf, BUF_SZ, fmt, ap);
    va_end(ap);

    if (_tty)
	fprintf(stderr, "%s: %s", progname, buf);

    /* read the string and replace any %'s with %% */
    replace_pc(buf);

    if (_log_syslog_init)
	syslog(LOG_ERR, buf);

    /* did the string exceed BUF_SZ? */
    if (ret > BUF_SZ - 1) {
	if (_tty)
	    fprintf(stderr, "WARNING: last log message exceeded log buffer "
		"size: %d\n", BUF_SZ - 1);
	if (_log_syslog_init)
	    syslog(LOG_ERR, "WARNING: last log message exceeded log buffer "
		"size: %d\n", BUF_SZ - 1);
    }

    exit(status);
}

/*
 *    function:	void replace_pc()
 * description:	read the string and replace any %'s with %%
 *     returns:	
 */
void
replace_pc (char *buf) {
    char	newbuf[BUF_SZ],
    		*next,
		*ptr = buf,
		*newptr = newbuf;
    int		total = 0,
		len;

    buf[BUF_SZ - 1] = '\0';

    if ((next = (char *) strstr(buf, "%")) == NULL) {
	return;
    }

    while (*ptr) {
	if ((next = (char *) strstr(ptr, "%")) == NULL) {
	    len = strlen(ptr);
	    if (total + len >= BUF_SZ) {
		len = BUF_SZ - 1 - total;
	    }
	    strncpy(newptr, ptr, len);
	    total += len;
	    break;
	}

	/* copy chars up to the next % */
	len = next - ptr;
	total += len;
	strncpy(newptr, ptr, len);
	newptr += len;
	ptr = next;

	/* insert %% */
	ptr++;
	strncpy(newptr, "%%", 2);
	total += 2;
	newptr += 2;
    }

    total++;
    newbuf[total] = '\0';
    strncpy(buf, newbuf, total);

    return;
}


/*
 *    function:	void start_syslog()
 * description:	initialize syslog
 *     returns:	
 */
void
start_syslog(char *ident, int logopt, int facility)
{
    if (debug & DEBUG_CALL || debug & DEBUG_INIT)
	debug_msg("start_syslog(%s, %d, %d)\n", ident, logopt, facility);

    if (_log_syslog_init)
	stop_syslog();

    /* switch to not print to stderr if not a tty */
    _tty = isatty(0);		/* HEAS: is 0 right fd? */

    openlog(ident, logopt, facility);

    _log_syslog_init = 1;

    return;
}

/*
 *    function:	void stop_syslog()
 * description:	close syslog
 *     returns:	
 */
void
stop_syslog(void)
{
    if (debug & DEBUG_CALL)
	debug_msg("stop_syslog()\n");

    if (_log_syslog_init)
	closelog();
    _log_syslog_init = 0;

    return;
}

/*
 *    function:	void debug_off()
 * description:	turn off debugging after SIGUSR1
 *       notes:	bison/yacc internal debugging can not be turned off once it
 *		is enabled.
 *     returns:	
 */
RETSIGTYPE
debug_off(void)
{
    if (debug & DEBUG_CALL)
	log_msg(LOG_DEBUG, "debug_off()\n");

    debug = 0;
    signal(SIGCHLD, SIG_IGN);

    return;
}
