/*
 * misc support functions
 */

#include "util.h"
#include "mem.h"

#include <ctype.h>

#if ! HAVE_SETENV
/*
 *  Subroutine:	setenv
 *       Usage:	int setenv()
 * Description:	emulate setenv() for platforms without
 *     returns:	non-zero on failure
 */
int
setenv(var, val, override)
    char		*var;
    char		*val;
    int			override;
{
    int			len = 0;
    char		*string;

    if (! var)
	return(1);
    len = strlen(var);

    /* if override is false (0) and the env var exists, setenv() fails */
    if (! override)
	if (getenv(var) != NULL)
	    return(-1);

    if (val)
	len += strlen(val);
    len += 2;
    if ((string = xmalloc(len)) == NULL) {
	errno = ENOMEM;
	return(-1);
    }

    if (snprintf(string, len, "%s=%s", var, val ? val : "") >= len) {
	free(string);
	return(-1);
    }

    len = putenv(string);
    free(string);

    return(len);
}
#endif

/*
 *  Subroutine:	strtohex
 *       Usage:	void strtohex()
 * Description:	convert a string to a hex string, 2 digits per character
 *     returns:	non-zero on failure
 */
int
strtohex(src, dst, dstlen)
    char	*src,
		*dst;
    int		dstlen;				/* length of dst buffer */
{

    if (dst == NULL)
	return(-1);
    if (src == NULL) {
	*dst = '\0';
	return(0);
    }

    if (dstlen < strlen(src) * 2 + 1)
	return(-1);

    while (*src != '\0') {
	sprintf(dst, "%02x", *src);
	dst += 2;
	src++;
    }
    dst = '\0';

    return(0);
}

/*
 *  Subroutine:	strtolower
 *       Usage:	void strtolower()
 * Description:	convert a string to lowercase
 *     returns:	non-zero on failure
 */
void
strtolower(str)
    char	*str;
{
    while (str != NULL && *str != '\0') {
	*str = tolower(*str);
	str++;
    }

    return;
}

#if ! HAVE_VASPRINTF
/*
 *  Subroutine:	asprintf
 *       Usage:	int asprintf()
 * Description:	emulate asprintf() for platforms without
 *     returns:	non-zero on malloc failure
 */
int
#if STDC_HEADERS
asprintf(char **ret, char *format, ...)
#else
asprintf(ret, format, va_alist)
    char	**ret;
    char	*format;
    va_dcl
#endif
{
    va_list	ap;
    int		len,
		vlen;

    if (ret == NULL)
	return(-1);

    if (format == NULL) {
	*ret = NULL;
	return(0);
    }

    len = strlen(format) + 1;
    if (ap != NULL)
	len = len * 2;

    if ((*ret = (char *) xmalloc(len)) == NULL)
	    return(-1);

    while (1) {
	va_start(ap, format);
	if ((vlen = vsnprintf(*ret, len, format, ap)) <= len) {
	    va_end(ap);
	    return(vlen);
	}
	va_end(ap);
	len = vlen + 1;
	if (xrealloc((void *)ret, len) == NULL) {
	    free(*ret);
	    *ret = NULL;
	    return(-1);
	}
    }

    /* NOTREACHED */
    return(0);
}

/*
 *  Subroutine:	vasprintf
 *       Usage:	int vasprintf()
 * Description:	emulate vasprintf() for platforms without
 *     returns:	non-zero on failure
 */
int
vasprintf(ret, format, ap)
    char	**ret;
    char	*format;
    va_list	ap;
{
    int		len,
		vlen;

    if (ret == NULL)
	return(-1);

    if (format == NULL) {
	*ret = NULL;
	return(0);
    }

    len = strlen(format) + 1;
    if (ap != NULL)
	len = len * 2;

    if ((*ret = (char *) xmalloc(len)) == NULL)
	    return(-1);

    while (1) {
	if ((vlen = vsnprintf(*ret, len, format, ap)) < len) {
	    return(vlen);
	}
	len = vlen + 1;
	if (xrealloc((void *)ret, len) == NULL) {
	    free(*ret);
	    *ret = NULL;
	    return(-1);
	}
    }

    /* NOTREACHED */
    return(0);
}
#endif
