/*
 * chksyslog -
 */
#include "main.h"
#include "version.h"
#include "mem.h"
#include "llist.h"
#include "log.h"
#include "util.h"

#include <signal.h>

extern int	debug;

#define SYSLCONF	"/etc/syslog.conf"

char	*progname,
	*cachepath,					/* fppos cache path */
	*homepath,					/* env(HOME) path */
	*tmppath,					/* env(TMPDIR) path */
	*tmpdir,
	*tmpu,
	*tmpv;
int	opt_i,
	read_stdin;
llist	cache_l, file_l, ignore_l, violate_l, join_l, skip_l;
FILE	*u,						/* tmpu */
	*v;						/* tmpv */

/* function prototypes */
int	main(int, char **);
void	Fprint(FILE *, char *);
void	ehandler(int);
int	fastfwd(FILE *, char *);
int	filter(FILE *, char *);
int	match(char *, llist *);
int	readcache(void);
void	read_filters(char *, llist *);
int	read_syslconf(FILE *, llist *);
int	savepos(void);
int	slng_readline(FILE *, char *, int);
int	updatepos(FILE *, char *);
void	usage(void);
void	vers(void);

int
main(int argc, char **argv)
{
    register int	op;
    char		*cp;
    extern char		*optarg;
    extern int		optind, opterr;
    char		*ignore, *violate = NULL, *join = NULL, *hostfile;
    char		hn[64],				/* hostname */
			*skip = NULL,			/* skip file */
			*tempfmt;			/* mktemp() format */
    FILE		*fp;
    int			append_OLD = 1,			/* -O */
			checkonly = 0;			/* -c */
    struct stat		sb;
    node		*next;

    if ((progname = strrchr(argv[0], '/')) != NULL)
	progname++;
    else
	progname = argv[0];

    /* parse command-line options */
    while ((op = getopt(argc, argv, "Ocdhisvf:")) != -1)
	switch (op) {
	    case 'O':			/* do append "OLD" to dirname() */
		append_OLD=0;
		break;
	    case 'c':			/* config check only */
		checkonly++;
		break;
	    case 'f':			/* skip file */
		skip = optarg;
		break;
	    case 'i':			/* incremental */
		++opt_i;
		if (cachepath == NULL) {
		    if ((cachepath = getenv("CHK_CACHE")) == NULL) {
		        if ((homepath = getenv("HOME")) == NULL) {
			    log_msg(LOG_WARNING, "WARNING: HOME not found in"
				"environment\n");
			}
		        if (asprintf(&cachepath, "%s%s%s",
				homepath == NULL ? "" : homepath,
				homepath == NULL ? "" : "/",
				CACHEFNAME) < 0) {
			    log_msgx(EX_OSERR, "malloc() failure: %s\n",
				strerror(errno));
			}
		    }
		}
		break;
	    case 'd':
		++debug;
		break;
	    case 's':			/* input is stdin */
		++read_stdin;
		break;
	    case 'v':
		vers();
		exit(EX_OK);
	    default:
		usage();
	}

    if (argc - optind < 1) {
	log_msg(LOG_ERR, "too few arguments\n");
	usage();
    }
    ignore = argv[optind++];
    if (optind < argc)
	violate = argv[optind++];
    if (optind < argc)
	join = argv[optind];

    /*
     * read the ignore, violate, join, and skip (and hostname specific
     * versions) files and compile the regexes
     */
    if (gethostname(hn, 64 - 1))
	log_msgx(EX_OSERR, "gethostname() failed: %s\n", strerror(errno));
    if ((hostfile = index(hn, '.')) != NULL)
	hostfile = '\0';
    read_filters(ignore, &ignore_l);
    if (asprintf(&hostfile, "%s.%s", ignore, hn) == -1)
	log_msgx(EX_OSERR, "malloc() failure: %s\n", strerror(errno));
    if (stat(hostfile, &sb) == 0)
	read_filters(hostfile, &ignore_l);
    free(hostfile);
    if (violate) {
	read_filters(violate, &violate_l);
	if (asprintf(&hostfile, "%s.%s", violate, hn) == -1)
	    log_msgx(EX_OSERR, "malloc() failure: %s\n", strerror(errno));
	if (stat(hostfile, &sb) == 0)
	    read_filters(hostfile, &violate_l);
	free(hostfile);
    }
    if (join) {
	read_filters(join, &join_l);
	if (asprintf(&hostfile, "%s.%s", violate, hn) == -1)
	    log_msgx(EX_OSERR, "malloc() failure: %s\n", strerror(errno));
	if (stat(hostfile, &sb) == 0)
	    read_filters(hostfile, &join_l);
	free(hostfile);
    }
    if (skip)
	read_filters(skip, &skip_l);

    /* -c.  if read_filters returned, the regexes compiled */
    if (checkonly)
	return(EX_OK);

    /* set umask to 077 */
    umask(S_IRWXG | S_IRWXO);

    /* figure out the temp dir path */
    if ((tmppath = getenv("TMPDIR")) == NULL)
	tmppath = TMPDIR;
    if (asprintf(&tempfmt, "%s/%s", tmppath, TMPDIRFMT) == -1)
	log_msgx(EX_OSERR, "malloc() failure: %s\n", strerror(errno));

    /* create temp dir */
    do {
	if ((tmpdir = mktemp(tempfmt)) == NULL)
	    log_msgx(EX_OSERR, "mktemp() failed: %s\n", strerror(errno));
	if (mkdir(tmpdir, 0700)) {
	    if (errno == EEXIST)
		continue;

	    log_msgx(EX_OSERR, "mkdir(%s) failed: %s\n", tmpdir,
		strerror(errno));
	} else
	    break;
    } while (1);

    /* cleanup on signal*/
    signal(SIGHUP, (void *) ehandler);
    signal(SIGINT, (void *) ehandler);
    signal(SIGQUIT, (void *) ehandler);
    signal(SIGTERM, (void *) ehandler);

    /* open output files */
    if (asprintf(&tmpu, "%s/%s", tmpdir, "unknown") < 0 ||
	asprintf(&tmpv, "%s/%s", tmpdir, "violations") < 0)
	log_msgx(EX_OSERR, "malloc() failure: %s\n", strerror(errno));
    if ((u = fopen(tmpu, "w+")) == NULL)
	log_msgx(EX_CANTCREAT, "fopen(%s) failure: %s\n", tmpu,
		strerror(errno));
    if ((v = fopen(tmpv, "w+")) == NULL)
	log_msgx(EX_CANTCREAT, "fopen(%s) failure: %s\n", tmpv,
		strerror(errno));

    if (read_stdin) {
	filter(stdin, "stdin");
    } else {
	/*
	 * read syslog.conf for filenames, then prepend OLD/ (if not -O)
	 * and append .0
	 */
	if (debug)
	    log_msg(LOG_DEBUG, "reading %s\n", SYSLCONF);
	if ((fp = fopen(SYSLCONF, "r")) == NULL)
	    log_msgx(EX_OSFILE, "open(%s) failed: %s\n", SYSLCONF,
		strerror(errno));
	read_syslconf(fp, &file_l);
	fclose(fp);

	/* filter each of the files found in syslog.conf */
	next = ll_next(&file_l, NULL);
	while (next) {
	    if (match((char *) next->data, &skip_l)) {
		if (debug)
		   log_msg(LOG_DEBUG, "skipping log file: %s\n",
			(char *) next->data);
	    } else {
		if (opt_i) {
		    if (asprintf(&hostfile, "%s", (char *) next->data) < 0)
			log_msgx(EX_OSERR, "malloc() failure: %s\n",
							strerror(errno));
		} else {
		    /* reformat the logfile name */
		    cp = rindex((char *) next->data, '/');
		    *cp = '\0'; cp++;
		    if (append_OLD) {
			if (asprintf(&hostfile, "%s/OLD/%s.0",
					(char *) next->data, cp) < 0) {
			    log_msgx(EX_OSERR, "malloc() failure: %s\n",
							strerror(errno));
			}
		    } else {
			if (asprintf(&hostfile, "%s/%s.0",
					(char *) next->data, cp) < 0) {
			    log_msgx(EX_OSERR, "malloc() failure: %s\n",
							strerror(errno));
			}
		    }
		}
		if (debug)
		   log_msg(LOG_DEBUG, "reading log file: %s\n", hostfile);
		if ((fp = fopen(hostfile, "r")) == NULL) {
		    log_msg(LOG_ERR, "open(%s) failed: %s\n",
			hostfile, strerror(errno));
		} else {
		    if (opt_i) {
			fastfwd(fp, hostfile);
			filter(fp, hostfile);
			updatepos(fp, hostfile);
		    } else
			filter(fp, hostfile);
		}
		free(hostfile);
	    }

	    next = ll_next(&file_l, next);
	}
    }

    fflush(u); rewind(u);
    fflush(v); rewind(v);
    if (stat(tmpv, &sb) == 0)
	if (sb.st_size > 0) {
	    printf("Violations:\n\n");
	    Fprint(v, tmpv);
	    printf("\n\n");
	}
    if (stat(tmpu, &sb) == 0)
	if (sb.st_size > 0) {
	    printf("Unknown logs:\n\n");
	    Fprint(u, tmpu);
	}

    /* save out file positions */
    if (opt_i)
	savepos();

    ehandler(0);
    return(0);
}

/*
 * description: copy file to stdout from current position to EOF
 *     returns:
 */
void
Fprint(FILE *fp, char *file)
{
     char	buf[8192];
     int	len;

     for (;;) {
	len = fread(buf, sizeof(char), 8191, fp);
	buf[len] = '\0';

	if (len)
	    fputs(buf, stdout);
	else
	    if (feof(fp))
		break;
	    else
		log_msg(LOG_ERR, "error reading %f: %s\n", file,
			strerror(errno));
     }

     return;
}

/*
 * description: exit/signal handler
 *     returns:
 */
void
ehandler(int sig)
{
    if (u)
	fclose(u);
    if (v)
	fclose(v);

    if (! debug) {
	if (tmpu)
	    unlink(tmpu);
	if (tmpv)
	    unlink(tmpv);
    }

    if (tmpdir)
	rmdir(tmpdir);

    exit(0);
}

/*
 * description: check cache file entry for *file against the current file and
 *		advance the file-pointer if appropriate.
 *     returns: non-zero on error
 */
int
fastfwd(FILE *fp, char *file)
{
    struct _fpos	*pos;
    struct stat		sb;
    node		*next = NULL;

    if (ll_empty(&cache_l))
	if (readcache())
	    log_msgx(EX_NOINPUT, "ERROR: cache file read failure: %s\n",
		strerror(errno));

    while ((next = ll_next(&cache_l, next)) != NULL) {
	if (next->data == NULL) {
	   ll_delete(&cache_l, next);
	   free(next);
	}
	pos = (struct _fpos *) next->data;

	if (strcmp(pos->fname, file) == 0) {
	    if (stat(file, &sb) != 0) {
		log_msg(LOG_ERR, "ERROR: stat(%s): %s\n", file,
			strerror(errno));
		return(1);
	    }

	    /* if the inode number matches and the size is >=, then ffwd,
	     * else delete the cache_l entry.
	     */
	    if (sb.st_ino == pos->st_ino && sb.st_size >= pos->st_size) {
		if (fseek(fp, pos->st_size, SEEK_SET) != 0) {
		    log_msg(LOG_ERR, "ERROR: fseek(%u) failed\n", pos->st_size);
		    return(1);
		}
	    } else {
		if ((next = ll_delete(&cache_l, next)) == NULL) {
		    log_msg(LOG_ERR, "ERROR: could not extract old file "
					"position structure\n", pos->st_size);
		    return(1);
		}
		if (next->data != NULL);
		    free(next->data);
		free(next);
	    }
	}
    }

    return(0);
}

/*
 * description: read a file and apply ignore, violations, and join regexes to
 *		each line until EOF.
 *     returns: non-zero on error
 */
int
filter(FILE *fp, char *file)
{
    char	buf[LINE_MAX*2],
		*bufp;

    bufp = buf;

/* XXX: need to make this does not throw away non-\n terminated lines in the
 * incremental case */
    do {
	if (fgets(buf, LINE_MAX, fp) == NULL) {
	    if (feof(fp))
		break;
	    log_msg(LOG_ERR, "error reading %s: %s\n", file, strerror(errno));
	    return(1);
	}
	bufp = buf + strlen(buf);

	if (match(buf, &join_l)) {
	    if (fgets(bufp, LINE_MAX, fp) == NULL) {
		if (feof(fp))
		    break;
		log_msg(LOG_ERR, "error reading %s: %s\n", file,
			strerror(errno));
		return(1);
	    }
	}

	if (match(buf, &violate_l)) {
	    if (fputs(buf, v) == EOF)
		log_msgx(EX_OSERR, "error writing %s: %s\n", tmpv,
			strerror(errno));
	} else if (! match(buf, &ignore_l)) {
	    if (fputs(buf, u) == EOF)
		log_msgx(EX_OSERR, "error writing %s: %s\n", tmpu,
			strerror(errno));
	}
	bufp = buf;
    } while (1);

    return(0);
}

/*
 * description: given a line, apply each regex in llist until there is a match
 *     returns: non-zero on match
 */
int
match(char *buf, llist *l)
{
    node	*next = l->head;

    next = ll_next(l, NULL);
    while (next) {
	if (!regexec(&(((struct _regex *)next->data)->preg), buf, 0, NULL, 0)) {
	    if (((struct _regex *)next->data)->negate)
		return(0);
	    return(1);
	}
	next = ll_next(l, next);
    }
    return(0);
}

/*
 * description: reads a position cache file into llist
 *     returns: non-zero on error
 */
int
readcache(void)
{
    FILE		*cfp;
    struct _fpos	*pos;

    /* it isnt an error if the cache file does not exist */
    if ((cfp = fopen(cachepath, "r")) == NULL) {
	if (errno == ENOENT)
	    return(0);
	else
	    return(1);
    }

    for (;;) {
	if ((pos = (struct _fpos *) xmalloc(sizeof(struct _fpos))) == NULL) {
	    log_msgx(EX_OSERR, "malloc() failure: %s\n", strerror(errno));
	    fclose(cfp);
	    return(1);
	}
	if (fread(pos, 1, sizeof(struct _fpos), cfp) != sizeof(struct _fpos)) {
	    if (feof(cfp)) {
		free(pos);
		break;
	    }
	    if (ferror(cfp))
		log_msg(LOG_ERR, "ERROR: error reading %s\n", cachepath);
	    else
		log_msg(LOG_ERR, "ERROR: error reading %s: short read\n",
			cachepath);
	    fclose(cfp);
	    free(pos);
	    return(1);
	}

	/* make sure the pathname is null terminated */
        pos->fname[PATH_MAX] = '\0';
	if (ll_append(&cache_l, pos) == NULL)
	    log_msgx(EX_OSERR, "malloc failed: %s\n", strerror(errno));
    }
    fclose(cfp);

    return(0);
}


/*
 * description: read a filter file, compile the regexes and append it to the
 *		list.
 *     returns: doesnt.  exits on error
 */
void
read_filters(char *file, llist *l)
{
    FILE		*fp;
    char		buf[LINE_MAX], *bufp;
    int			err,
			line = 0;
    struct _regex	*r;

    if (file == NULL)
	return;

    if ((fp = fopen(file, "r")) == NULL)
	log_msgx(EX_NOINPUT, "open(%s) failed: %s\n", file, strerror(errno));

    do {
	if (fgets(buf, LINE_MAX, fp) == NULL) {
	    if (feof(fp))
		break;
	    log_msgx(EX_IOERR, "error reading %s line %d: %s\n", file, line,
		strerror(errno));
	}
	line++;

	/* skip blank and comments lines */
	if (*buf == '\0' || *buf == '#' || *buf == '\n')
	    continue;

	/* strip the return */
	err = strlen(buf) - 1;
	if (buf[err] == '\n')
	    buf[err] = '\0';

	if (strlen(buf) == LINE_MAX -1 && buf[LINE_MAX-1] != '\n')
	    log_msgx(EX_DATAERR, "%s: line %d exeeds max length of %d\n",
		file, line, LINE_MAX);

	if ((r = xmalloc(sizeof(struct _regex))) == NULL)
	    log_msgx(EX_OSERR, "malloc failed: %s\n", strerror(errno));

	bufp = buf;
	if (buf[0] == '!') {
	    r->negate = 1;
	    bufp++;
	} else if (buf[0] == '\\' && buf[1] == '!')
	    bufp++;
	if ((err = regcomp(&(r->preg), bufp,
			   REG_EXTENDED | REG_NOSUB | REG_NEWLINE))) {
	    regerror(err, &(r->preg), buf, LINE_MAX);
	    log_msgx(EX_DATAERR, "regex compile failed %s:%d: %s\n", file,
		line, buf);
	}
	if (ll_append(l, r) == NULL)
	    log_msgx(EX_OSERR, "malloc failed: %s\n", strerror(errno));
    } while (1);

    fclose(fp);
    return;
}

/*
 * description: read syslog.conf
 *     returns: non-zero on error
 */
int
read_syslconf(FILE *fp, llist *l)
{
    char	buf[LINE_MAX];
#define NREGEXES	5
    char	*r[NREGEXES] = {
			/* comment or blank line */
			"^[[:space:]]*(#.*)?$",
			/* BSD filename */
			"^[[:space:]]*[^[:space:]]+[[:space:]]+(/[^[:space:]]+)[[:space:]]*$",
			/* syslog-ng destination line */
			"^[[:space:]]*destination [^[:space:]]+",
			/* syslog-ng destination filename */
			"file\\(\\\"([^[:space:]\"]+)\\\"",
			/* % END % */
			0
			},
		*file;
    regex_t	p[NREGEXES];
    regmatch_t	pmatch[2];
    int		c,
		err,
		line = 0,
		style = 0;
	
    /* compile regexes */
    memset(p, 0, sizeof(regex_t) * NREGEXES);
    for (c = 0; r[c] != NULL; c++) {
	if ((err = regcomp(&p[c], r[c], REG_EXTENDED))) {
	    regerror(err, &p[c], buf, LINE_MAX);
	    if (debug) {
		log_msg(LOG_ERR, "internal regex compile failed: %s\n", buf);
		abort();
	    } else
		log_msgx(EX_SOFTWARE, "internal regex compile failed: %s\n",
			buf);
	}
    }

    do {
	if (fgets(buf, LINE_MAX, fp) == NULL) {
	    if (feof(fp))
		break;
	    log_msgx(EX_IOERR, "error reading syslog.conf: %s\n",
		strerror(errno));
	}
	line++;

	/* skip blank and comments lines */
	if (!regexec(&p[0], buf, 0, NULL, 0))
	    continue;

	/* look for log filename */
	file = NULL;
	if (style >= 0  && !regexec(&p[2], buf, 0, NULL, 0)) {
	    /* syslog-ng style */
	    style++;
	    if ((err = slng_readline(fp, buf, LINE_MAX)) < 0) {
		log_msgx(LOG_ERR, "premature EOF reading syslog.conf\n");
	    } else {
		line += err;
		if (!regexec(&p[3], buf, 2, pmatch, 0)) {
		    file = buf + pmatch[1].rm_so;
		    *(buf + pmatch[1].rm_eo) = '\0';
		}
	    }
	} else if (style <= 0 && !regexec(&p[1], buf, 2, pmatch, 0)) {
	    /* BSD style */
	    style--;
	    file = buf + pmatch[1].rm_so;
	    *(buf + pmatch[1].rm_eo) = '\0';
	}
	if (file) {
	    if (strncmp(file, "/dev/", 5)) {
		if (ll_append(l, strdup(file)) == NULL)
		    log_msgx(EX_OSERR, "malloc() failure: %s\n",
			strerror(errno));
		if (debug)
		    log_msg(LOG_DEBUG, "syslog.conf file: %s\n", file);
	    } else
		if (debug)
		    log_msg(LOG_DEBUG, "skipping syslog.conf file: %s\n", file);
	   file = NULL;
	}
    } while (1);

    return(0);
}

/*
 * description: save cache_l to cache file
 *     returns: non-zero on error
 */
int
savepos(void)
{
    FILE		*cfp;
    struct _fpos	*pos;
    node		*next = NULL;

    if (ll_empty(&cache_l))
	return(0);

    if ((cfp = fopen(cachepath, "w+")) == NULL)
	log_msgx(LOG_ERR, "ERROR: open(%s) failed: %s\n", cachepath,
		strerror(errno));

    while ((next = ll_next(&cache_l, next)) != NULL) {
	pos = (struct _fpos *) next->data;

	if (fwrite(pos, 1, sizeof(struct _fpos), cfp) != sizeof(struct _fpos)) {
	    if (feof(cfp)) {
		free(pos);
		break;
	    }
	    if (ferror(cfp))
		log_msg(LOG_ERR, "ERROR: error writing %s\n", cachepath);
	    else
		log_msg(LOG_ERR, "ERROR: error writing %s: short write\n",
			cachepath);
	    fclose(cfp);
	    free(pos);
	    return(1);
	}
    }
    fclose(cfp);

    return(0);
}

/*
 * description: complete a syslog-ng destination clause
 *     returns: less than zero on error, else number of CRs
 */
int
slng_readline(FILE *fp, char *buf, int s)
{
    int		brace = 0,
		obrace = 0,
		cr = 0;
    char	*cp = buf;

    for (;;) {
	switch (*cp) {
	case '{':
	    obrace++;
	    brace++;
	    break;
	case '}':
	    brace--;
	    break;
	case '\n':
	    cr--;
	    break;
	case '\0':
	    if (cp - buf == s || (*cp = fgetc(fp)) == EOF)
		return(-1);
	    *(cp + 1) = '\0';
	    continue;
	    break;
	}

	if (obrace && brace == 0)
	    break;
	cp++;
    }

    return(cr);
}

/*
 * description: update fp position struct in cache_l
 *     returns: non-zero on error
 */
int
updatepos(FILE *fp, char *file)
{
    struct _fpos	*pos;
    struct stat		sb;
    node		*next = NULL;

    if (ll_empty(&cache_l))
	if (readcache())
	    log_msgx(EX_NOINPUT, "ERROR: cache file read failure: %s\n",
		strerror(errno));

    /* look for *file in cache_l */
    while ((next = ll_next(&cache_l, next)) != NULL) {
	pos = (struct _fpos *) next->data;

	if (strcmp(pos->fname, file) == 0) {
	    pos->st_size = ftell(fp);
	    return(0);
	}
    }

    /*
     * if we reached the end of the cache_l, we didnt find an entry for *file.
     * so create one
     */
    if (next == NULL) {
	if ((pos = (struct _fpos *) xmalloc(sizeof(struct _fpos))) == NULL) {
	    log_msgx(EX_OSERR, "malloc() failure: %s\n", strerror(errno));
	    return(1);
	}

	/* make sure the pathname is null terminated */
        strncpy(pos->fname, file, PATH_MAX);
        pos->fname[PATH_MAX] = '\0';
	pos->st_size = ftell(fp);
    }

    /* record the inode */
    if (stat(file, &sb) != 0) {
	log_msg(LOG_ERR, "ERROR: stat(%s): %s\n", file, strerror(errno));
	free(pos->fname);
	free(pos);
	return(1);
    }
    pos->st_ino = sb.st_ino;

    if (ll_append(&cache_l, pos) == NULL)
	log_msgx(EX_OSERR, "malloc failed: %s\n", strerror(errno));

    return(0);
}

void
usage(void)
{
    fprintf(stderr,
"usage: %s [-hv] [-i | s] [-f <skip file>] <ignore file> [<violations file>] [<join file>]\n"
"       -O  do not insert OLD/ in logfile path\n"
"       -c  read regex files, compile, then exit non-zero on error\n"
"       -f  specify chksyslog.skip filename\n"
"       -i  incremental check\n"
"       -s  read stdin\n"
"       -h  print this help msg\n"
"       -v  version info\n",
	 progname);
    exit(EX_USAGE);
}

void
vers(void)
{
    fprintf(stderr,
"%s version %s, Copyright (C) 2001 heas@shrubbery.net\n"
"The %s package is free software which you are welcome to redistribute\n"
"under certain conditions and comes with ABSOLUTELY NO WARRANTY; for details\n"
"see the COPYING file that came with the distribution.\n",
	progname, version, package);
    exit(EX_USAGE);
}
