/*
 * malloc/memory functions
 */

#include <config.h>
#include <mem.h>
#include <log.h>

/*
 * description:	retuns ptr to zero'd block of size bytes, else logs msg
 *		and returns null
 */
void *
xmalloc(int size)
{
    void	*ptr;

    if (debug & DEBUG_CALL)
	debug_msg("xmalloc(%d)\n", size);

    if (size <= 0) {
	log_msg(LOG_ERR, "ERROR: xmalloc call with insane value %d bytes "
		"(%s, %d)\n", size, __FILE__, __LINE__);
	return(NULL);
    }

    if ((ptr = malloc(size)) == NULL) {
	log_msg(LOG_ERR, "ERROR: malloc failed for %d bytes (%s, %d): %s\n",
		size, __FILE__, __LINE__, strerror(errno));
	return(NULL);
    }

    /* zero the space */
    memset(ptr, 0, size);

    return(ptr);
}

/*
 * description:	adjusts ptr to size bytes, else logs msg and returns null
 *     returns:	returns NULL on malloc error
 */
void *
xrealloc(void **ptr, int size)
{
    void	*newptr;

    if (debug & DEBUG_CALL)
	debug_msg("xrealloc(%d)\n", size);

    if (ptr == NULL) {
	log_msg(LOG_ERR, "ERROR: xrealloc call with null pointer %x "
		"(%s, %d)\n", ptr, __FILE__, __LINE__);
	return(NULL);
    }

    if (*ptr == NULL) {
	/* HEAS: NULL ptr should just call xmalloc */
	log_msg(LOG_ERR, "ERROR: xrealloc call with null pointer %x "
		"(%s, %d)\n", ptr, __FILE__, __LINE__);
	return(NULL);
    }

    if (size <= 0) {
	log_msg(LOG_ERR, "ERROR: xrealloc call with insane value %d bytes "
		"(%s, %d)\n", size, __FILE__, __LINE__);
	return(NULL);
    }

    if ((newptr = realloc(*ptr, size)) == NULL) {
	log_msg(LOG_ERR, "ERROR: malloc failed for %d bytes (%s, %d): %s\n",
		size, __FILE__, __LINE__, strerror(errno));
	return(NULL);
    }
    *ptr = newptr;

    return(newptr);
}
