/*-
 * Copyright (c) 1988, 1989, 1992, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 * Copyright (c) 1996 BBN, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] =
"@(#) Copyright (c) 1988, 1989, 1992, 1993, 1994\n\
	The Regents of the University of California.  All rights reserved.\n\
@(#) Copyright (c) 1996 BBN Inc. All rights reserved.\n";
#endif /* not lint */

#ifndef lint
/* from: static char sccsid[] = "@(#)rshd.c	8.2 (Berkeley) 4/6/94"; */
static const char rcsid[] = "@(#) $Id: rshd.c,v 1.3 2003/04/12 17:11:46 heas Exp $";
#endif /* not lint */

/*
 * remote shell server:
 *	[port]\0
 *	remuser\0
 *	locuser\0
 *	command\0
 *	data
 */

#include "config.h"

#include <sys/param.h>
#if HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/socket.h>


#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#if HAVE_NETINET_IN_SYSTM_H
# include <netinet/in_systm.h>
#endif
#include <netinet/ip.h>


#include <errno.h>
#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_STRING_H
# include <string.h>
#endif

# include <syslog.h>

#if HAVE_UNISTD_H
# include <unistd.h>
#endif
#if HAVE_GRP_H
# include <grp.h>
#endif

#include "rcpd.h"

#ifndef _PATH_TTY
#define _PATH_TTY "/dev/tty"
#endif

int	keepalive = 1;
int	check_all;
int	sent_null;

struct sockaddr_in from;

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct linger linger;
	int ch, on = 1, fromlen;

#if POSIXLY_CORRECT
	/* make gnu-geopt() behave */
	putenv("POSIXLY_CORRECT=1");
#endif

	openlog("rcpd", LOG_PID | LOG_ODELAY, LOG_DAEMON);

	opterr = 0;
	while ((ch = getopt(argc, argv, "an")) != EOF)
		switch (ch) {
		case 'a':
			check_all = 1;
			break;
		case 'n':
			keepalive = 0;
			break;
		case '?':
		default:
			usage();
			break;
		}

	argc -= optind;
	argv += optind;

	fromlen = sizeof (from);
	if (getpeername(0, (struct sockaddr *)&from, &fromlen) < 0) {
		syslog(LOG_ERR, "getpeername: %m");
		_exit(1);
	}
	if (keepalive &&
	    setsockopt(0, SOL_SOCKET, SO_KEEPALIVE, (char *)&on,
	    sizeof(on)) < 0)
		syslog(LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");
	linger.l_onoff = 1;
	linger.l_linger = 60;			/* XXX */
	if (setsockopt(0, SOL_SOCKET, SO_LINGER, (char *)&linger,
	    sizeof (linger)) < 0)
		syslog(LOG_WARNING, "setsockopt (SO_LINGER): %m");
#ifdef IP_TOS
	/*
	 * This is sort of imperfect as the SYN openning the connection will
	 * be sent with whatever TOS/Precedence inetd chooses to use, which
	 * is most likely IPTOS_PREC_ROUTINE, though perhaps ought to be the
	 * same as what the incoming SYN had.
	 */

	{
	  int tos;

	  tos = IPTOS_PREC_INTERNETCONTROL;
	  if (setsockopt(0, IPPROTO_IP, IP_TOS, (char*)&tos, sizeof(tos)) < 0)
	    syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
	}
#endif
	doit(&from);
	/* NOTREACHED */
	return 1;
}

void 
doit(fromp)
	struct sockaddr_in *fromp;
{
	struct hostent *hp;
	u_short port;
	fd_set ready, readfrom;
	int cc, nfd, pv[2], pid, s;
#if DONT_HAVE_O_NONBLOCK
	int one = 1;
#endif
	char *hostname, *errorstr, *errorhost;
	char sig, buf[BUFSIZ];
	char cmdbuf[NCARGS+1], locuser[32], remuser[32];
	char remotehost[2 * MAXHOSTNAMELEN + 1];

	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);

#ifdef DEBUG
	{ int t = open(_PATH_TTY, O_RDWR);
	  if (t >= 0) {
		ioctl(t, TIOCNOTTY, (char *)0);
		(void) close(t);
	  }
	}
#endif

	fromp->sin_port = ntohs((u_short)fromp->sin_port);
	if (fromp->sin_family != AF_INET) {
		syslog(LOG_ERR, "malformed \"from\" address (af %d)\n",
		    fromp->sin_family);
		exit(1);
	}
#ifdef IP_OPTIONS
      {
	u_char optbuf[BUFSIZ/3], *cp;
	char lbuf[BUFSIZ], *lp;
	int optsize = sizeof(optbuf), ipproto;
	struct protoent *ip;

	if ((ip = getprotobyname("ip")) != NULL)
		ipproto = ip->p_proto;
	else
		ipproto = IPPROTO_IP;
	if (!getsockopt(0, ipproto, IP_OPTIONS, (char *)optbuf, &optsize) &&
	    optsize != 0) {
		lp = lbuf;
		for (cp = optbuf; optsize > 0; cp++, optsize--, lp += 3)
			sprintf(lp, " %2.2x", *cp);
		syslog(LOG_NOTICE,
		    "Connection received from %s using IP options (ignored):%s",
		    inet_ntoa(fromp->sin_addr), lbuf);
		if (setsockopt(0, ipproto, IP_OPTIONS,
		    (char *)NULL, optsize) != 0) {
			syslog(LOG_ERR, "setsockopt IP_OPTIONS NULL: %m");
			exit(1);
		}
	}
      }
#endif

/*
 * XXX
 * Not needed for rcpd - we don't care about what port the user is
 * coming from.
 *	--asp@partan.com
 *	Thu Feb  3 21:39:42 EST 2000
 *
 *		if (fromp->sin_port >= IPPORT_RESERVED ||
 *		    fromp->sin_port < IPPORT_RESERVED/2) {
 *			syslog(LOG_NOTICE|LOG_AUTH,
 *			    "Connection from %s on illegal port %u",
 *			    inet_ntoa(fromp->sin_addr),
 *			    fromp->sin_port);
 *			exit(1);
 *		}
 *************/

	(void) alarm(60);
	port = 0;
	for (;;) {
		char c;
		if ((cc = read(STDIN_FILENO, &c, 1)) != 1) {
			if (cc < 0)
				syslog(LOG_NOTICE, "read: %m");
			shutdown(0, 1+1);
			exit(1);
		}
		if (c== 0)
			break;
		port = port * 10 + c - '0';
	}

	(void) alarm(0);
	if (port != 0) {
		int lport = IPPORT_RESERVED - 1;
		s = rresvport(&lport);
		if (s < 0) {
			syslog(LOG_ERR, "can't get stderr port: %m");
			exit(1);
		}
			if (port >= IPPORT_RESERVED) {
				syslog(LOG_ERR, "2nd port not reserved\n");
				exit(1);
			}
#ifdef IP_TOS
		{
		  int tos;

		  tos = IPTOS_PREC_INTERNETCONTROL;
		  if (setsockopt(s, IPPROTO_IP, IP_TOS, (char*)&tos, sizeof(tos)) < 0)
		    syslog(LOG_WARNING, "setsockopt 2nd port (IP_TOS): %m");
		}
#endif

		fromp->sin_port = htons(port);
		if (connect(s, (struct sockaddr *)fromp, sizeof (*fromp)) < 0) {
			syslog(LOG_INFO, "connect second port %d: %m", port);
			exit(1);
		}
	}

#ifdef notdef
	/* from inetd, socket is already on 0, 1, 2 */
	dup2(f, 0);
	dup2(f, 1);
	dup2(f, 2);
#endif
	errorstr = NULL;
	hp = gethostbyaddr((char *)&fromp->sin_addr, sizeof (struct in_addr),
		fromp->sin_family);
	if (hp) {
		/*
		 * If name returned by gethostbyaddr is in our domain,
		 * attempt to verify that we haven't been fooled by someone
		 * in a remote net; look up the name and check that this
		 * address corresponds to the name.
		 */
		hostname = hp->h_name;
		if (check_all || local_domain(hp->h_name)) {
			strncpy(remotehost, hp->h_name, sizeof(remotehost) - 1);
			remotehost[sizeof(remotehost) - 1] = 0;
			errorhost = remotehost;
			hp = gethostbyname(remotehost);
			if (hp == NULL) {
				syslog(LOG_INFO,
				    "Couldn't look up address for %s",
				    remotehost);
				errorstr =
				"Couldn't look up address for your host (%s)\n";
				hostname = inet_ntoa(fromp->sin_addr);
			} else for (; ; hp->h_addr_list++) {
				if (hp->h_addr_list[0] == NULL) {
					syslog(LOG_NOTICE,
					  "Host addr %s not listed for host %s",
					    inet_ntoa(fromp->sin_addr),
					    hp->h_name);
					errorstr =
					    "Host address mismatch for %s\n";
					hostname = inet_ntoa(fromp->sin_addr);
					break;
				}
				if (!bcmp(hp->h_addr_list[0],
				    (caddr_t)&fromp->sin_addr,
				    sizeof(fromp->sin_addr))) {
					hostname = hp->h_name;
					break;
				}
			}
		}
		strcpy(buf, hostname);
		strcat(buf, " (");
		strcat(buf, inet_ntoa(fromp->sin_addr));
		strcat(buf, ")");
		hostname = buf;
	} else
		errorhost = hostname = inet_ntoa(fromp->sin_addr);

	getstr(remuser, sizeof(remuser), "remuser");
	getstr(locuser, sizeof(locuser), "locuser");
	getstr(cmdbuf, sizeof(cmdbuf), "command");

	syslog(LOG_INFO|LOG_AUTH, "%s@%s as %s: cmd='%.80s'",
	       remuser, hostname, locuser, cmdbuf);

	if (strncmp(cmdbuf, "rcp ", 4) != 0) {
	  syslog(LOG_INFO|LOG_AUTH,
		 "tried to run non-rcp cmd '%s', proceding",
		 cmdbuf);
#ifdef notdef
	  error("Won't run non-rcp commands.\n");
	  exit(1);
#endif
	}

	(void) write(STDERR_FILENO, "\0", 1);
	sent_null = 1;

	if (port) {
		if (pipe(pv) < 0) {
			error("Can't make pipe.\n");
			exit(1);
		}
		pid = fork();
		if (pid == -1)  {
			error("Can't fork; try again.\n");
			exit(1);
		}
		if (pid) {
			(void) close(0);
			(void) close(1);
			(void) close(2);
			(void) close(pv[1]);

			FD_ZERO(&readfrom);
			FD_SET(s, &readfrom);
			FD_SET(pv[0], &readfrom);
			if (pv[0] > s)
				nfd = pv[0];
			else
				nfd = s;

#if DONT_HAVE_O_NONBLOCK	/* XXX - configure should define something */
				ioctl(pv[0], FIONBIO, (char *)&one);
#else
				fcntl(pv[0], F_SETFL, O_NONBLOCK);
#endif

			/* should set s nbio! */
			nfd++;
			do {
				ready = readfrom;
					if (select(nfd, &ready, (fd_set *)0,
					  (fd_set *)0, (struct timeval *)0) < 0)
						break;
				if (FD_ISSET(s, &ready)) {
					int	ret;
						ret = read(s, &sig, 1);
					if (ret <= 0)
						FD_CLR(s, &readfrom);
					else
						killpg(pid, sig);
				}
				if (FD_ISSET(pv[0], &ready)) {
					errno = 0;
					cc = read(pv[0], buf, sizeof(buf));
					if (cc <= 0) {
						shutdown(s, 1+1);
						FD_CLR(pv[0], &readfrom);
					} else {
							(void)
							  write(s, buf, cc);
					}
				}
			} while (FD_ISSET(s, &readfrom) ||
			    FD_ISSET(pv[0], &readfrom));
			exit(0);
		}
#if SETPGRP_VOID
		setpgrp();
#else
		setpgrp(0, getpid());
#endif
		(void) close(s);
		(void) close(pv[0]);
		dup2(pv[1], 2);
		close(pv[1]);
	}

	{
	  struct passwd *p;

	  if ((p = getpwnam("nobody")) == (struct passwd*) NULL) {
	    syslog(LOG_ERR, "getpwnam(nobody): %m");
	    exit(1);
	  }

	  if (setgid(p->pw_gid)) {
		syslog(LOG_ERR, "setgid(): %m");
		exit(1);
	  }
	  if (initgroups("nobody", p->pw_gid)) {
	    syslog(LOG_ERR, "initgroups(): %m");
	    exit(1);
	  }
	  if (setuid(p->pw_uid)) {
	    syslog(LOG_ERR, "setuid(): %m");
	    exit(1);
	  }
	}
	
	{
	  int rcpargc = 0;
	  char *rcpargv[10];

	  rcpargv[rcpargc] = strtok(cmdbuf, " ");
	  while ((rcpargv[++rcpargc] = strtok(NULL, " ")) && (rcpargc < 9))
		;
	  rcp(rcpargc, rcpargv);
	}

	exit(1);
}

/*
 * Report error to client.  Note: can't be used until second socket has
 * connected to client, or older clients will hang waiting for that
 * connection first.
 */
#if defined(__STDC__) || defined(STDC_HEADERS)
#include <stdarg.h>
#else
#include <varargs.h>
#endif

void
#if defined(__STDC__) || defined(STDC_HEADERS)
error(const char *fmt, ...)
#else
error(fmt, va_alist)
	char *fmt;
        va_dcl
#endif
{
	va_list ap;
	int len;
	char *bp, buf[BUFSIZ];
#if defined(__STDC__) || defined(STDC_HEADERS)
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	bp = buf;
	if (sent_null == 0) {
		*bp++ = 1;
		len = 1;
	} else
		len = 0;

	vsnprintf(bp, BUFSIZ - len, fmt, ap);
	va_end(ap);

	syslog(LOG_ERR, "returning error: %s",buf);
	(void)write(STDERR_FILENO, buf, len + strlen(bp));
}

void
getstr(buf, cnt, err)
	char *buf, *err;
	int cnt;
{
	char c;

	do {
		if (read(STDIN_FILENO, &c, 1) != 1)
			exit(1);
		*buf++ = c;
		if (--cnt == 0) {
			error("%s too long\n", err);
			exit(1);
		}
	} while (c != 0);
}

/*
 * Check whether host h is in our local domain,
 * defined as sharing the last two components of the domain part,
 * or the entire domain part if the local domain has only one component.
 * If either name is unqualified (contains no '.'),
 * assume that the host is local, as it will be
 * interpreted as such.
 */
int
local_domain(h)
	char *h;
{
	char localhost[MAXHOSTNAMELEN];
	char *p1, *p2;

	localhost[0] = 0;
	(void) gethostname(localhost, sizeof(localhost));
	p1 = topdomain(localhost);
	p2 = topdomain(h);
	if (p1 == NULL || p2 == NULL || !strcasecmp(p1, p2))
		return (1);
	return (0);
}

char *
topdomain(h)
	char *h;
{
	char *p, *maybe = NULL;
	int dots = 0;

	for (p = h + strlen(h); p >= h; p--) {
		if (*p == '.') {
			if (++dots == 2)
				return (p);
			maybe = p;
		}
	}
	return (maybe);
}

void
usage()
{

	syslog(LOG_ERR, "usage: rcpd [-an]");
	exit(2);
}
